#!/usr/bin/env python
__author__ = 'Felipe Reyes <felipe.reyes@canonical.com>'

import re
import os
import shutil
import subprocess
import sys

from charmhelpers.core.hookenv import (
    config,
    local_unit,
    log,
    relation_ids,
    relation_get,
    relation_set,
    unit_get,
    Hooks,
    UnregisteredHookError,
)
from charmhelpers.core import templating
from charmhelpers.core.host import (
    restart_on_change,
    service_reload,
    service_start,
    service_stop,
)
from charmhelpers.fetch import apt_install, apt_update
from charmhelpers.contrib.network import ufw
import memcached_utils


DOT = os.path.dirname(os.path.abspath(__file__))
ETC_DEFAULT_MEMCACHED = '/etc/default/memcached'
ETC_MEMCACHED_CONF = '/etc/memcached.conf'
ETC_MUNIN_NODE_CONF = '/etc/munin/munin-node.conf'
LOCAL_NAGIOS_PLUGINS = '/usr/local/lib/nagios/plugins'
MEMCACHE_NAGIOS_PLUGIN = '/usr/local/lib/nagios/plugins/check_memcache.py'
RESTART_MAP = {ETC_MEMCACHED_CONF: ['memcached'],
               ETC_MUNIN_NODE_CONF: ['munin-node']}
hooks = Hooks()


class MemcachedError(Exception):
    pass


@hooks.hook('install')
def install():
    apt_update(fatal=True)
    apt_install(["memcached", "python-cheetah", "python-memcache"], fatal=True)

    with open(ETC_DEFAULT_MEMCACHED, 'w') as f:
        f.write('ENABLE_MEMCACHED=yes\n')

    ufw.enable()
    ufw.service('ssh', 'open')


@hooks.hook('start')
def start():
    service_start('memcached')


@hooks.hook('stop')
def stop():
    service_stop('memcached')


@hooks.hook('cache-relation-joined')
def cache_relation_joined():

    settings = {'host': unit_get('private-address'),
                'port': config('tcp-port'),
                'udp-port': config('udp-port')}

    for rid in relation_ids('cache'):
        relation_set(rid, **settings)

    addr = relation_get('private-address')
    if addr:
        log('Granting memcached access to {}'.format(addr), level='INFO')
        memcached_utils.grant_access(addr)


@hooks.hook('cache-relation-departed')
def cache_relation_departed():
    addr = relation_get('private-address')
    log('Revoking memcached access to {}'.format(addr))
    memcached_utils.revoke_access(addr)


@hooks.hook('config-changed')
@restart_on_change(RESTART_MAP)
def config_changed():
    mem_size = config('size')

    if mem_size == 0:
        output = subprocess.check_output(['free', '-m'])
        mem_size = int(re.findall(r'\d+', output.split('\n')[2])[1])
        mem_size = int(mem_size * 0.9)

    configs = {'mem_size': mem_size,
               'large_pages_enabled': False}

    for key in ['request-limit', 'min-item-size', 'slab-page-size', 'threads',
                'disable-auto-cleanup', 'disable-cas', 'factor',
                'connection-limit', 'tcp-port', 'udp-port']:
        configs[key.replace('-', '_')] = config(key)
        log('Config: {} = {}'.format(key, configs[key.replace('-', '_')]),
            level='DEBUG')

    if mem_size >= 2048:
        disable_lp = config('disable-large-pages')
        if not disable_lp:
            log('Memory >= 2GB, using large-pages to speed up access',
                level='INFO')
            save_hp = int(subprocess.check_output(['sysctl', '-n',
                                                   'vm.nr_hugepages']))

            npages = mem_size / 2
            subprocess.check_output(['sysctl', '-w',
                                     'vm.nr_hugepages={}'.format(npages)])
            allocated = int(subprocess.check_output(['sysctl', '-n',
                                                     'vm.nr_hugepages']))

            if allocated < npages:
                log(('Cannot allocate $npages contiguous pages for huge '
                     'page allocation'), level='WARNING')
                log('Setting vm.nr_hugepages back to {}'.format(save_hp),
                    level='WARNING')
                subprocess.check_output(['sysctl', '-w',
                                         'vm.nr_hugepages={}'.format(save_hp)])
                log('Will disable large pages and fall back to regular memory',
                    level='WARNING')
                configs['large_pages_enabled'] = False
            else:
                log(('Allocated $npages contiguous blocks only '
                     'for huge pages.'.format(npages)), level='WARNING')
                configs['large_pages_enabled'] = True

    templating.render('memcached.conf', ETC_MEMCACHED_CONF, configs)

    # In case port changed, inform consumers
    cache_relation_joined()


@hooks.hook('munin-relation-changed')
@restart_on_change(RESTART_MAP)
def munin_relation_changed():
    remote_ip = relation_get('private-address')

    if not remote_ip:
        log('Remote node must provide IP', level='INFO')
        sys.exit(0)

    munin_server_ip = memcached_utils.munin_format_ip(remote_ip)

    # make sure munin port is open, the access is restricted by munin
    ufw.service('4949', 'open')

    # make sure it's instaled
    apt_update(fatal=True)
    apt_install(['munin-node'], fatal=True)
    configs = {'munin_server': munin_server_ip}
    templating.render('munin-node.conf', ETC_MUNIN_NODE_CONF, configs)
    relation_set(ip=unit_get('private-address'))


@hooks.hook('nrpe-external-master-relation-changed')
@restart_on_change(RESTART_MAP)
def nrpe_external_master_relation_changed():

    # make sure it's installed
    apt_update(fatal=True)
    apt_install(['nagios-nrpe-server'], fatal=True)
    ufw.service('5666', 'open')

    if not os.path.isdir(LOCAL_NAGIOS_PLUGINS):
        os.makedirs(LOCAL_NAGIOS_PLUGINS)

    if not os.path.isfile(MEMCACHE_NAGIOS_PLUGIN):
        shutil.copy(os.path.join(DOT, "..", "files", "nrpe-external-master",
                                 "check_memcache.py"),
                    os.path.join(LOCAL_NAGIOS_PLUGINS, "check_memcache.py"))
        os.chmod(os.path.join(LOCAL_NAGIOS_PLUGINS, "check_memcache.py"),
                 0o555)

    nagios_hostname = "{}-{}".format(config('nagios_context'),
                                     local_unit().replace('/', '-'))
    configs = {'tcp_port': config('tcp_port'),
               'servicegroup': config('nagios-context'),
               'hostname': nagios_hostname}

    templating.render('nrpe_check_memcached.cfg.tmpl',
                      '/etc/nagios/nrpe.d/check_memcached.cfg',
                      configs)
    templating.render('nrpe_export.cfg.tmpl',
                      ('/var/lib/nagios/export/service__{}_check_'
                       'memcached.cfg').format(nagios_hostname),
                      configs)

    service_reload('nagios-nrpe-server')


@hooks.hook('upgrade-charm')
def upgrade_charm():
    install()
    config_changed()


def main():
    try:
        hooks.execute(sys.argv)
    except UnregisteredHookError as e:
        log('Unknown hook {} - skipping.'.format(e))


if __name__ == '__main__':
    main()
