__author__ = 'Felipe Reyes <felipe.reyes@canonical.com>'

import mock
import os
import shutil
import tempfile
from test_utils import CharmTestCase
import memcached_hooks


DOT = os.path.dirname(os.path.abspath(__file__))
TO_PATCH = [
    'apt_install',
    'apt_update',
    'service_start',
    'service_stop',
    'relation_ids',
    'relation_set',
    'relation_get',
    'unit_get',
    'config',
    'log',
]
FREE_MEM_SMALL = """             total       used       free     shared    \
buffers     cached
Mem:          2002        744       1257          0         82        433
-/+ buffers/cache:        227       1774
Swap:            0          0          0
"""
FREE_MEM_BIG = """             total       used       free     shared    \
buffers     cached
Mem:         24010      13371      10639        207        484       6669
-/+ buffers/cache:       6216      17793
Swap:         7811          0       7811
"""


@mock.patch('charmhelpers.contrib.network.ufw.is_enabled', lambda: True)
@mock.patch('charmhelpers.core.hookenv.log', mock.MagicMock())
class TestMemcachedHooks(CharmTestCase):
    def setUp(self):
        self.fake_repo = {}
        super(TestMemcachedHooks, self).setUp(memcached_hooks, TO_PATCH)
        self.relation_get.return_value = '127.0.0.1'
        self.tmpdir = tempfile.mkdtemp()
        memcached_hooks.ETC_DEFAULT_MEMCACHED = os.path.join(self.tmpdir,
                                                             "memcached")
        memcached_hooks.ETC_MEMCACHED_CONF = os.path.join(self.tmpdir,
                                                          "memcached.conf")

    def tearDown(self):
        super(TestMemcachedHooks, self).tearDown()
        shutil.rmtree(self.tmpdir, ignore_errors=True)

    @mock.patch('subprocess.check_output')
    def test_install(self, check_output):
        memcached_hooks.install()

        self.apt_update.assert_called_with(fatal=True)
        self.apt_install.assert_called_with(["memcached", "python-cheetah",
                                             "python-memcache"], fatal=True)
        check_output.assert_any_call(['ufw', 'allow', 'ssh'])

    def test_start(self):
        memcached_hooks.start()
        self.service_start.assert_called_with('memcached')

    def test_stop(self):
        memcached_hooks.stop()
        self.service_stop.assert_called_with('memcached')

    @mock.patch('subprocess.check_output')
    @mock.patch('memcached_hooks.relation_get')
    @mock.patch('memcached_utils.grant_access')
    @mock.patch('memcached_utils.config')
    @mock.patch('memcached_utils.log')
    def test_cache_relation_joined(self, log, config, grant_access,
                                   relation_get, check_output):
        configs = {'tcp-port': '1234', 'udp-port': '3456'}

        def f(c):
            return configs.get(c, None)

        self.config.side_effect = f
        config.side_effect = f
        relation_get.return_value = '127.0.1.1'
        self.unit_get.return_value = '127.0.0.1'
        self.relation_ids.return_value = ['cache:1', 'cache:2']
        memcached_hooks.cache_relation_joined()
        self.relation_set.assert_any_call('cache:1', **{'host': '127.0.0.1',
                                                        'port': '1234',
                                                        'udp-port': '3456'})
        self.relation_set.assert_any_call('cache:2', **{'host': '127.0.0.1',
                                                        'port': '1234',
                                                        'udp-port': '3456'})
        grant_access.assert_called_with('127.0.1.1')

    @mock.patch('subprocess.check_output')
    @mock.patch('memcached_hooks.relation_get')
    @mock.patch('memcached_utils.config')
    @mock.patch('memcached_utils.log')
    @mock.patch('charmhelpers.contrib.network.ufw.revoke_access')
    def test_cache_relation_departed(self, revoke_access, log, config,
                                     relation_get, check_output):
        config.return_value = 1234
        relation_get.return_value = '127.0.1.1'
        memcached_hooks.cache_relation_departed()

        revoke_access.assert_any_call('127.0.1.1', port='1234', proto='tcp')
        revoke_access.assert_any_call('127.0.1.1', port='1234', proto='udp')

    @mock.patch('subprocess.Popen')
    @mock.patch('subprocess.check_output')
    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('memcached_utils.config')
    @mock.patch('memcached_utils.log')
    def test_config_changed_defaults(self, log, config, render, check_output,
                                     popen):
        p = mock.Mock()
        p.configure_mock(**{'communicate.return_value': ('stdout', 'stderr'),
                            'returncode': 0})
        popen.return_value = p

        configs = self.test_config.config

        def f(c):
            return configs.get(c, None)

        self.config.side_effect = f
        config.side_effect = f

        memcached_hooks.config_changed()

        passed_vars = render.call_args[0][2]

        self.assertEqual(passed_vars['mem_size'],
                         self.test_config.config['size'])

    @mock.patch('subprocess.Popen')
    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('subprocess.check_output')
    @mock.patch('memcached_utils.log')
    def test_config_changed_size_set_0_small(self, log, check_output, render,
                                             popen):
        p = mock.Mock()
        p.configure_mock(**{'communicate.return_value': ('stdout', 'stderr'),
                            'returncode': 0})
        popen.return_value = p

        configs = {'size': 0}

        def f(c):
            return configs.get(c, None)

        self.config.side_effect = f

        def g(*args, **kwargs):
            if args[0] == ['free', '-m']:
                return FREE_MEM_SMALL
            else:
                return ""

        check_output.side_effect = g

        memcached_hooks.config_changed()
        passed_vars = render.call_args[0][2]

        self.assertEqual(passed_vars['mem_size'], 1596)
        self.assertFalse(passed_vars['large_pages_enabled'])

    @mock.patch('subprocess.Popen')
    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('subprocess.check_output')
    @mock.patch('memcached_utils.log')
    def test_config_changed_size_set_0_big(self, log, check_output, render,
                                           popen):
        p = mock.Mock()
        p.configure_mock(**{'communicate.return_value': ('stdout', 'stderr'),
                            'returncode': 0})
        popen.return_value = p

        configs = {'size': 0,
                   'disable-large-pages': False}

        def f(c):
            return configs.get(c, None)

        self.config.side_effect = f

        def g(*args, **kwargs):
            if args[0] == ['free', '-m']:
                return FREE_MEM_BIG
            if args[0] == ['sysctl', '-n', 'vm.nr_hugepages']:
                return int(16013 / 2)
            else:
                return ""

        check_output.side_effect = g

        memcached_hooks.config_changed()
        passed_vars = render.call_args[0][2]

        self.assertEqual(passed_vars['mem_size'], 16013)
        check_output.assert_any_call(['sysctl', '-w',
                                      'vm.nr_hugepages={}'.format(16013/2)])
        self.assertTrue(passed_vars['large_pages_enabled'])

    @mock.patch('subprocess.Popen')
    @mock.patch('charmhelpers.core.templating.render')
    @mock.patch('subprocess.check_output')
    @mock.patch('memcached_utils.log')
    def test_config_changed_size_failed_set_hpages(self, log, check_output,
                                                   render, popen):
        p = mock.Mock()
        p.configure_mock(**{'communicate.return_value': ('stdout', 'stderr'),
                            'returncode': 0})
        popen.return_value = p

        configs = {'size': 0,
                   'disable-large-pages': False}

        def f(c):
            return configs.get(c, None)

        self.config.side_effect = f

        def g(*args, **kwargs):
            if args[0] == ['free', '-m']:
                return FREE_MEM_BIG
            if args[0] == ['sysctl', '-n', 'vm.nr_hugepages']:
                return 16
            else:
                return ""

        check_output.side_effect = g

        memcached_hooks.config_changed()
        passed_vars = render.call_args[0][2]

        self.assertEqual(passed_vars['mem_size'], 16013)
        self.assertFalse(passed_vars['large_pages_enabled'])
        check_output.assert_any_call(['sysctl', '-w',
                                      'vm.nr_hugepages={}'.format(16013/2)])
        check_output.assert_any_call(['sysctl', '-w',
                                      'vm.nr_hugepages=16'])

    @mock.patch('os.fchown')
    @mock.patch('os.chown')
    @mock.patch('memcached_utils.log')
    @mock.patch('subprocess.Popen')
    @mock.patch('memcached_utils.config')
    @mock.patch('subprocess.check_output')
    @mock.patch('charmhelpers.contrib.network.ufw.enable')
    @mock.patch('charmhelpers.contrib.network.ufw.service')
    @mock.patch('charmhelpers.core.hookenv.charm_dir')
    @mock.patch('charmhelpers.core.host.log')
    def test_upgrade_charm(self, log, charm_dir, service, enable, check_output,
                           config, popen, *args):
        p = mock.Mock()
        p.configure_mock(**{'communicate.return_value': ('stdout', 'stderr'),
                            'returncode': 0})
        popen.return_value = p

        charm_dir.return_value = os.path.join(DOT, '..')
        config.return_value = 12111
        memcached_hooks.upgrade_charm()
        self.apt_install.assert_any_call(['memcached', 'python-cheetah',
                                          'python-memcache'], fatal=True)
        enable.assert_called_with()
        service.assert_called_with('ssh', 'open')
